# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Fast pareto frontier computation using Jax."""
import functools

import jax
from jax import numpy as jnp
import jaxtyping as jt
import numpy as np
from vizier._src.pyvizier.multimetric import pareto_optimal


def _is_dominated(
    y1: jt.Float[jt.Array, "M"],
    y2: jt.Float[jt.Array, "M"],
    strict: bool = True,
) -> jt.Bool[jt.Array, ""]:
  """True if y2 > y1 (or y2 >= y1 if strict is False) every coordinate."""
  dominated_or_equal = jnp.all(y1 <= y2)
  if strict:
    return dominated_or_equal & jnp.any(y2 > y1)
  else:
    return dominated_or_equal


@functools.partial(jax.jit, static_argnames="strict")
def _is_pareto_optimal_against(
    yy: jt.Float[jt.Array, "B1 M"],
    baseline: jt.Float[jt.Array, "B2 M"],
    *,
    strict: bool,
) -> jt.Bool[jt.Array, "B1"]:
  """Computes if nothing in `baseline` dominates `yy`.

  Args:
    yy: array of shape [B1, M] where M is number of metrics.
    baseline: array of shape [B2, M] where M is number of metrics.
    strict: If true, strict dominance is used.

  Returns:
    Boolean array of shape [B1]
  """
  jax_dominated_mv = jax.vmap(
      functools.partial(_is_dominated, strict=strict), (None, 0), 0
  )  #  ([b,a], [a]) -> [b]
  jax_dominated_mm = jax.vmap(jax_dominated_mv, (0, None),
                              0)  #  ([b,a], [c,a]) -> [b,c]
  return jnp.logical_not(jnp.any(jax_dominated_mm(yy, baseline),
                                 axis=-1))  # N_y, N_b -> N_y


def is_frontier(
    ys: jt.Float[jt.ArrayLike, "B M"],
    *,
    num_shards: int = 10,
    verbose: bool = False,
) -> jt.Bool[jt.ArrayLike, "B"]:
  """Efficiently compute `_is_pareto_optimal_against(ys, ys, strict=True)`.

  Divide `ys` into shards and gradually trim down the candidates.

  Args:
    ys: Array of shape [B, M] where M is number of metrics.
    num_shards: Each sharding results in filtering, i.e. indexing the array with
      boolean vector. This operation can be very expensive and dominate the cost
      of computation. Use a moderate number sublinear in B, e.g. log(B).
    verbose:

  Returns:
    Boolean numpy Array of shape [B].
  """
  idx = np.linspace(0, ys.shape[0], num_shards).astype(np.int32)
  idx = list(reversed(idx))
  # Initialize candidates with all points
  frontier = np.ones(ys.shape[0], dtype=np.bool_)

  for begin, end in zip(idx[1:], idx[:-1]):
    candidates = ys[frontier]
    # Filter candidates by comparing against the slice.
    if verbose:
      print(f"Compare {len(candidates)} against {begin}:{end}.")
    tt = _is_pareto_optimal_against(candidates, ys[begin:end], strict=True)
    frontier[frontier] = tt
  return frontier


class JaxParetoOptimalAlgorithm(pareto_optimal.BaseParetoOptimalAlgorithm):
  """Jax-based functions for calculating pareto frontiers."""

  def is_pareto_optimal(self, points: np.ndarray) -> np.ndarray:
    """Jax-enabled Pareto frontier algorithm. See base class."""
    return np.array(is_frontier(points), dtype=bool)

  def is_pareto_optimal_against(self, points: np.ndarray,
                                dominating_points: np.ndarray,
                                strict: bool) -> np.ndarray:
    """Jax-enabled optimality/domination algorithm. See base class."""
    return np.array(
        _is_pareto_optimal_against(points, dominating_points, strict=strict),
        dtype=bool,
    )


def get_frontier(
    ys: jt.Float[jt.ArrayLike, "B M"],
    *,
    num_shards: int = 10,
    verbose: bool = True,
) -> jnp.ndarray:
  """Efficiently compute `ys[_is_pareto_optimal_against(ys, ys, strict=True)]` using iterative filtering.

  Divide `ys` into shards and gradually trim down the candidates.
  `get_frontier` doesn't call `is_frontier`, because `get_frontier` runs faster
  by not slicing the full `ys` every iteration.

  Args:
    ys: Array of shape [B, M] where M is number of metrics.
    num_shards: Each sharding results in filtering, i.e. indexing the array with
      boolean vector. This operation can be very expensive and dominate the cost
      of computation. Use a moderate number sublinear in B, e.g. log(B).
    verbose:

  Returns:
    Array of shape [B, M].
  """
  idx = np.linspace(0, ys.shape[0], num_shards).astype(np.int32)
  idx = list(reversed(idx))
  # Initialize candidates with all points
  candidates = jnp.asarray(ys)
  for begin, end in zip(idx[1:], idx[:-1]):
    # Filter candidates by comparing against the slice.
    if verbose:
      # Use print. This method won't run in production anyways.
      print(f"Compare {len(candidates)} against {begin}:{end}.")
    tt = _is_pareto_optimal_against(candidates, ys[begin:end], strict=True)
    candidates = candidates[tt]
  return candidates


@jax.jit
def pareto_rank(ys: jt.Float[jt.ArrayLike, "B M"]) -> jt.Int[jt.ArrayLike, "B"]:
  """Returns the pareto rank."""
  jax_dominated_mv = jax.vmap(
      functools.partial(_is_dominated, strict=True), (None, 0), 0
  )  #  ([b,a], [a]) -> [b]
  jax_dominated_mm = jax.vmap(
      jax_dominated_mv, (0, None), 0
  )  #  ([b,a], [c,a]) -> [b,c]
  domination_matrix = jax_dominated_mm(ys, ys)
  return jnp.sum(domination_matrix, axis=1)


def _cum_hypervolume_origin(
    points: jt.Float[jt.ArrayLike, "B M"], vector: jt.Float[jt.Array, "... M"]
) -> jt.Float[jt.Array, "B"]:
  """Returns a randomized approximation of the cumulative dominated hypervolume.

  See Section 3, Lemma 5 of https://arxiv.org/pdf/2006.04655.pdf for a fuller
  explanation of the technique. This assumes the reference point is the
  origin.

  NOTE: This returns an unnormalized hypervolume.

  Args:
    points: Any set of points with shape (num_points, dimension).
    vector: A vector of length dimension.

  Returns:
    Approximated cumulative dominated hypervolume of points[:i]. Length is
    num_points.
  """
  ratios = points / vector
  coordinate_min_ratio = jnp.min(ratios, axis=1)
  return jax.lax.cummax(coordinate_min_ratio, axis=0)**len(vector)


@jax.jit
def jax_cum_hypervolume_origin(
    points: jt.Float[jt.ArrayLike, "B M"], vectors: jt.Float[jt.Array, "B2 M"]
) -> jt.Float[jt.Array, "B B2"]:
  #  ([B,M], [B2,M]) -> [B,B2]
  cum_hypervolume_mm = jax.vmap(_cum_hypervolume_origin, (None, 0), 0)
  return jnp.mean(cum_hypervolume_mm(points, vectors), axis=0)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for noisy_experimenter."""

import numpy as np

from vizier import pyvizier
from vizier._src.benchmarks.experimenters import noisy_experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob

from absl.testing import absltest
from absl.testing import parameterized


class NoisyExperimenterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('Sphere', bbob.Sphere), ('Rastrigin', bbob.Rastrigin),
      ('BuecheRastrigin', bbob.BuecheRastrigin),
      ('LinearSlope', bbob.LinearSlope),
      ('AttractiveSector', bbob.AttractiveSector),
      ('StepEllipsoidal', bbob.StepEllipsoidal),
      ('RosenbrockRotated', bbob.RosenbrockRotated), ('Discus', bbob.Discus),
      ('BentCigar', bbob.BentCigar), ('SharpRidge', bbob.SharpRidge),
      ('DifferentPowers', bbob.DifferentPowers),
      ('Weierstrass', bbob.Weierstrass), ('SchaffersF7', bbob.SchaffersF7),
      ('SchaffersF7IllConditioned', bbob.SchaffersF7IllConditioned),
      ('GriewankRosenbrock', bbob.GriewankRosenbrock),
      ('Schwefel', bbob.Schwefel), ('Katsuura', bbob.Katsuura),
      ('Lunacek', bbob.Lunacek), ('Gallagher101Me', bbob.Gallagher101Me))
  def testDeterministicNoiseApply(self, func):
    dim = 2
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim))
    noisy_exptr = noisy_experimenter.NoisyExperimenter(
        exptr=exptr, noise_fn=lambda v: v - 1)

    parameters = exptr.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t = pyvizier.Trial(parameters={
        param.name: float(index) for index, param in enumerate(parameters)
    })

    exptr.evaluate([t])
    metric_name = exptr.problem_statement().metric_information.item().name
    unnoised_value = t.final_measurement_or_die.metrics[metric_name].value

    noisy_exptr.evaluate([t])
    noised_value = t.final_measurement_or_die.metrics[metric_name].value
    self.assertEqual(unnoised_value - 1, noised_value)
    self.assertEqual(
        unnoised_value,
        t.final_measurement_or_die.metrics[metric_name + '_before_noise'].value,
    )

  @parameterized.named_parameters(
      ('NO_NOISE', 'NO_NOISE', 1e-5),
      ('SEVERE_ADDITIVE_GAUSSIAN', 'SEVERE_ADDITIVE_GAUSSIAN', 3),
      ('MODERATE_ADDITIVE_GAUSSIAN', 'MODERATE_ADDITIVE_GAUSSIAN', 0.3),
      ('LIGHT_ADDITIVE_GAUSSIAN', 'LIGHT_ADDITIVE_GAUSSIAN', 0.03),
      ('MODERATE_GAUSSIAN', 'MODERATE_GAUSSIAN', 0.05),
      ('SEVERE_GAUSSIAN', 'SEVERE_GAUSSIAN', 0.5),
      ('MODERATE_UNIFORM', 'MODERATE_UNIFORM', 0.2),
      ('SEVERE_UNIFORM', 'SEVERE_UNIFORM', 3.5),
      ('MODERATE_SELDOM_CAUCHY', 'MODERATE_SELDOM_CAUCHY', 10.3),
      ('SEVERE_SELDOM_CAUCHY', 'SEVERE_SELDOM_CAUCHY', 100.3),
  )
  def testGaussianNoiseApply(self, noise: str, delta: float):
    dim = 2
    exptr = numpy_experimenter.NumpyExperimenter(
        bbob.Sphere, bbob.DefaultBBOBProblemStatement(dim))
    noisy_exptr = noisy_experimenter.NoisyExperimenter.from_type(
        exptr=exptr, noise_type=noise
    )

    parameters = exptr.problem_statement().search_space.parameters
    t = pyvizier.Trial(parameters={
        param.name: float(index) for index, param in enumerate(parameters)
    })

    exptr.evaluate([t])
    metric_name = exptr.problem_statement().metric_information.item().name
    unnoised_value = t.final_measurement_or_die.metrics[metric_name].value

    noisy_exptr.evaluate([t])
    noised_value1 = t.final_measurement_or_die.metrics[metric_name].value

    noisy_exptr.evaluate([t])
    noised_value2 = t.final_measurement_or_die.metrics[metric_name].value

    # Seldom noise is only injected sporadically.
    if 'SELDOM' not in noise and noise != 'NO_NOISE':
      self.assertNotEqual(noised_value1, noised_value2)
    self.assertAlmostEqual(noised_value1, unnoised_value, delta=delta)
    self.assertAlmostEqual(noised_value2, unnoised_value, delta=delta)

  def testSeedDeterminism(self):
    dim = 2
    seed = 7
    exptr = numpy_experimenter.NumpyExperimenter(
        bbob.Sphere, bbob.DefaultBBOBProblemStatement(dim)
    )
    noisy_exptr = noisy_experimenter.NoisyExperimenter.from_type(
        exptr=exptr, noise_type='SEVERE_UNIFORM', seed=seed
    )

    parameters = exptr.problem_statement().search_space.parameters
    t = pyvizier.Trial(
        parameters={
            param.name: float(index) for index, param in enumerate(parameters)
        }
    )
    metric_name = exptr.problem_statement().metric_information.item().name

    noise_value_sequence = []
    for _ in range(10):
      noisy_exptr.evaluate([t])
      noise_value_sequence.append(
          t.final_measurement_or_die.metrics[metric_name].value
      )

    # Global NP seed should not affect randomness.
    np.random.seed(0)

    noisy_exptr = noisy_experimenter.NoisyExperimenter.from_type(
        exptr=exptr, noise_type='SEVERE_UNIFORM', seed=seed
    )
    noise_value_sequence_after = []
    for _ in range(10):
      noisy_exptr.evaluate([t])
      noise_value_sequence_after.append(
          t.final_measurement_or_die.metrics[metric_name].value
      )
    self.assertSequenceAlmostEqual(
        noise_value_sequence, noise_value_sequence_after
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Uses a predictor as a surrogate model for evaluations."""
import copy
from typing import Sequence

import jax
from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter


class PredictorExperimenter(experimenter.Experimenter):
  """Uses predictor as a surrogate for objective value."""

  def __init__(
      self,
      predictor: vza.Predictor,
      problem_statement: vz.ProblemStatement,
      seed: int = 0,
  ):
    """Init.

    Args:
      predictor: Surrogate model for extrapolating unseen trials.
      problem_statement: Original problem statement for suggestions.
      seed: RNG seed for predictor.
    """
    self._predictor = predictor
    self._problem_statement = problem_statement
    self._rng = jax.random.PRNGKey(seed)
    self._objective_name = self._problem_statement.single_objective_metric_name

  def evaluate(self, suggestions: Sequence[vz.Trial]):
    prediction = self._predictor.predict(suggestions, self._rng)
    for i, suggestion in enumerate(suggestions):
      evaluation = prediction.mean[i]
      suggestion.complete(
          vz.Measurement(metrics={self._objective_name: evaluation})
      )

  def problem_statement(self) -> vz.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def __repr__(self) -> str:
    return (
        f'PredictorExperimenter on problem {self._problem_statement} with'
        f' {self._predictor}'
    )



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Experimenter that discretizes the parameters of search space."""

import copy
from typing import Mapping, Sequence, Union

import numpy as np
from vizier import pyvizier
from vizier._src.benchmarks.experimenters import experimenter
from vizier.pyvizier import converters


class DiscretizingExperimenter(experimenter.Experimenter):
  """DiscretizingExperimenter discretizes the parameters of search space."""

  def __init__(
      self,
      exptr: experimenter.Experimenter,
      discretization: Mapping[str, pyvizier.MonotypeParameterSequence],
      *,
      allow_oov: bool = False,
  ):
    """DiscretizingExperimenter discretizes continuous parameters.

    Currently only supports flat double search spaces. Note that the discretized
    parameters must fit within the bounds of the continuous parameters. This
    also supports CATEGORICAL parameters but feasible categories must be
    convertible to floats.

    Args:
      exptr: Underlying experimenter to be wrapped.
      discretization: Dict of parameter name to discrete/categorical values.
      allow_oov: Allows out of vocabulary values for Trial parameter in
        Evaluate. If True, evaluate the underlying experimenter at any given
        parameter values, whether feasible or not.

    Raises:
      ValueError: Non-double underlying parameters or discrete values OOB.
    """
    self._exptr = exptr
    self._discretization = discretization
    self._allow_oov = allow_oov
    exptr_problem_statement = exptr.problem_statement()

    if exptr_problem_statement.search_space.is_conditional:
      raise ValueError(
          'Search space should not have conditional'
          f' parameters  {exptr_problem_statement}'
      )

    search_params = exptr_problem_statement.search_space.parameters
    param_names = [param.name for param in search_params]
    for name in discretization.keys():
      if name not in param_names:
        raise ValueError(
            f'Parameter {name} not in search space'
            f' parameters for discretization: {search_params}'
        )

    self._problem_statement = copy.deepcopy(exptr_problem_statement)
    self._problem_statement.search_space = pyvizier.SearchSpace()
    for parameter in search_params:
      if parameter.name not in discretization:
        self._problem_statement.search_space.add(parameter)
        continue

      if parameter.type != pyvizier.ParameterType.DOUBLE:
        raise ValueError(
            f'Non-double parameters cannot be discretized {parameter}'
        )
      # Discretize the parameters.
      min_value, max_value = parameter.bounds
      for value in discretization[parameter.name]:
        float_value = float(value)
        if float_value > max_value or float_value < min_value:
          raise ValueError(f'Discretized values are not in bounds {parameter}')
      self._problem_statement.search_space.add(
          pyvizier.ParameterConfig.factory(
              name=parameter.name,
              feasible_values=discretization[parameter.name],
              scale_type=parameter.scale_type,
              external_type=parameter.external_type,
          )
      )

  def problem_statement(self) -> pyvizier.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def evaluate(self, suggestions: Sequence[pyvizier.Trial]) -> None:
    """Evaluate the trials after conversion to double."""

    old_parameters = []
    for suggestion in suggestions:
      old_parameters.append(suggestion.parameters)
      new_parameter_dict = {}
      for name, param in suggestion.parameters.items():
        if name in self._discretization:
          if self._allow_oov:
            if param.value not in self._discretization[name]:
              raise ValueError(
                  f'Parameter {param} not in {self._discretization[name]}'
              )
          new_parameter_dict[name] = param.as_float
        else:
          new_parameter_dict[name] = param
      suggestion.parameters = pyvizier.ParameterDict(new_parameter_dict)
    self._exptr.evaluate(suggestions)
    for old_param, suggestion in zip(old_parameters, suggestions):
      suggestion.parameters = old_param

  def __repr__(self):
    return f'DiscretizingExperimenter({self._discretization}) on {self._exptr}'

  @classmethod
  def create_with_grid(
      cls,
      exptr: experimenter.Experimenter,
      grid_discretization_count: Mapping[str, int],
      convert_to_str: Union[bool, Mapping[str, bool]] = False,
  ) -> 'DiscretizingExperimenter':
    """Creates Experimenter with continuous parameters discretized via grid.

    Note that the grid is generated according to ModelInputConverter scaling.

    Args:
      exptr: Experimenter with continuous parameters to be discretized.
      grid_discretization_count: Mapping from parameter names to number of
        feasible values to generate on grid. Parameter names should be a subset
        of all parameter names in the search space of the exptr.
      convert_to_str: For each parameter, if true, convert values to
        categories/strings. Otherwise, the values are discrete numbers. Can also
        be pure boolean, for all parameters.

    Returns:
      DiscreteExperimenter.

    Raises:
      ValueError: Invalid keys, discretizing non-double/singleton parameters.
    """
    if isinstance(convert_to_str, bool):
      convert_to_str = {
          k: convert_to_str for k in grid_discretization_count.keys()
      }

    if set(grid_discretization_count.keys()) != set(convert_to_str.keys()):
      raise ValueError(
          f'Grid discretization keys {grid_discretization_count.keys()} must'
          f' match convert_to_str keys {convert_to_str.keys()}'
      )

    # Create grid for each parameter config.
    discretization = {}
    problem_statement = copy.deepcopy(exptr.problem_statement())
    if not set(grid_discretization_count.keys()).issubset(
        set(problem_statement.search_space.parameter_names)
    ):
      raise ValueError(
          f'Grid discretization keys {grid_discretization_count.keys()}'
          f' are not in search space {problem_statement.search_space}'
      )

    for param in problem_statement.search_space.parameters:
      if param.name in grid_discretization_count:
        if param.type != pyvizier.ParameterType.DOUBLE:
          raise ValueError(
              f'Non-double parameters cannot be grid-discretized {param}'
          )

        # Grid creation logic for param.
        num_feasible_points = grid_discretization_count[param.name]
        min_value, max_value = param.bounds

        if min_value == max_value:
          raise ValueError(f'Cannot discretize singleton parameter {param}')
        converter = converters.DefaultModelInputConverter(param, scale=True)
        grid_scalars = np.linspace(0.0, 1.0, num=num_feasible_points)
        grid_values = converter.to_parameter_values(grid_scalars)

        if convert_to_str[param.name]:
          discretization[param.name] = [
              point.as_str for point in grid_values if point is not None
          ]
        else:
          discretization[param.name] = [
              point.as_float for point in grid_values if point is not None
          ]
        if len(discretization[param.name]) < len(grid_values):
          raise ValueError(
              '%d None values in grid values'
              % (len(grid_values) - len(discretization))
          )

    return cls(exptr, discretization)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""CMA-ES designer."""
import json
import queue
from typing import Optional, Sequence

from evojax.algo import cma_jax
import jax.numpy as jnp
import numpy as np

from vizier import algorithms as vza
from vizier import pyvizier as vz
from vizier.pyvizier import converters
from vizier.utils import json_utils


class CMAESDesigner(vza.PartiallySerializableDesigner):
  """CMA-ES designer wrapping evo-jax.

  NOTE: Since the base version of CMA-ES expects the entire population size to
  be evaluated before an update, we must use temporary queues to hold partially
  finished populations.
  """

  def __init__(self, problem_statement: vz.ProblemStatement, **cma_kwargs):
    """Init.

    Args:
      problem_statement: Must use a flat DOUBLE-only search space.
      **cma_kwargs: Keyword arguments for the CMA_ES_JAX class.
    """
    self._problem_statement = problem_statement
    self._metric_name = self._problem_statement.metric_information.item().name

    self._search_space = self._problem_statement.search_space
    if self._search_space.is_conditional:
      raise ValueError(
          f'This designer {self} does not support conditional search.')
    for parameter_config in self._search_space.parameters:
      if not parameter_config.type.is_continuous():
        raise ValueError(
            f'This designer {self} only supports continuous parameters.')
    self._num_params = len(self._search_space.parameters)
    if self._num_params < 2:
      raise ValueError(
          'CMA-ES only supports search spaces with >=2 parameters. Current'
          f' number of parameters: {self._num_params}'
      )

    # CMA-ES expects a maximization problem by default, so we flip signs for
    # minimization metrics.
    self._converter = converters.TrialToArrayConverter.from_study_config(
        self._problem_statement,
        scale=True,
        flip_sign_for_minimization_metrics=True,
    )
    self._cma_es_jax = cma_jax.CMA_ES_JAX(
        param_size=self._num_params, **cma_kwargs)
    self._trial_population = queue.Queue(
        maxsize=self._cma_es_jax.hyper_parameters.pop_size)

  def update(
      self, completed: vza.CompletedTrials, all_active: vza.ActiveTrials
  ) -> None:
    completed_trials = list(completed.trials)

    # Keep inserting completed trials into population. If population is full,
    # a CMA-ES update and queue clear are triggered.
    while completed_trials:
      self._trial_population.put(completed_trials.pop())

      if self._trial_population.full():
        # Once full, make a full CMA-ES update.
        features, labels = self._converter.to_xy(
            list(self._trial_population.queue))
        # CMA-ES expects fitness to be shape (pop_size,) and solutions of shape
        # (pop_size, num_params).
        self._cma_es_jax.tell(
            fitness=jnp.array(labels[:, 0]), solutions=jnp.array(features))
        self._trial_population.queue.clear()

  def suggest(self,
              count: Optional[int] = None) -> Sequence[vz.TrialSuggestion]:
    """Make new suggestions.

    Args:
      count: Makes best effort to generate this many suggestions. If None,
        suggests as many as the algorithm wants.

    Returns:
      New suggestions.
    """
    count = count or 1
    cma_suggestions = np.array(self._cma_es_jax.ask(count))

    # Convert CMA suggestions to suggestions.
    return [
        vz.TrialSuggestion(params)
        for params in self._converter.to_parameters(cma_suggestions)
    ]

  def load(self, metadata: vz.Metadata) -> None:
    cma_state = json.loads(
        metadata.ns('cma')['state'], object_hook=json_utils.numpy_hook)
    self._cma_es_jax.load_state(cma_state)

  def dump(self) -> vz.Metadata:
    cma_state = self._cma_es_jax.save_state()
    metadata = vz.Metadata()
    metadata.ns('cma')['state'] = json.dumps(
        cma_state, cls=json_utils.NumpyEncoder)
    return metadata



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Libraries for testing different Tuners on the shared backend."""
# pylint:disable=protected-access
# pylint:disable=invalid-name
import inspect
import threading
from typing import Type

from absl import flags
from absl import logging
import pyglove as pg
from vizier._src.pyglove import backend

from absl.testing import absltest

FLAGS = flags.FLAGS


class RandomAlgorithm(pg.DNAGenerator):
  """Dummy algorithm for testing."""

  def setup(self, dna_spec: pg.DNASpec):
    super().setup(dna_spec)
    self._random = pg.geno.Random(seed=1)
    self._random.setup(dna_spec)

  def _propose(self) -> pg.DNA:
    return self._random.propose()


@pg.members([
    ('max_trial_id', pg.typing.Int()),
    ('max_step', pg.typing.Int()),
])
class SizeLimitingStoppingPolicy(pg.tuning.EarlyStoppingPolicy):
  """Policy that stops trials when either ID or step is larger."""

  def _on_bound(self):
    super()._on_bound()
    self.requested_trial_steps = []
    self.stopped_trial_steps = []

  def should_stop_early(self, trial: pg.tuning.Trial) -> bool:
    if trial.status == 'COMPLETED':
      return False

    measurement = trial.measurements[-1]
    self.requested_trial_steps.append((trial.id, measurement.step))
    should_stop = (
        trial.id > self.max_trial_id or measurement.step > self.max_step
    )
    if should_stop:
      self.stopped_trial_steps.append((trial.id, measurement.step))
    return should_stop


class VizierTest(absltest.TestCase):
  """Base class for Vizier-based tests."""

  def __init__(
      self,
      backend_class: Type[backend.VizierBackend], *args, **kwargs
  ):
    self._backend_class = backend_class
    super().__init__(*args, **kwargs)


class SampleTest(VizierTest):
  """Tests for PyGlove sampling with Vizier backend."""

  def setUp(self):
    super().setUp()
    self._backend_class.use_study_prefix(self.id())

  def __init__(
      self,
      backend_class: Type[backend.VizierBackend],
      *args,
      builtin_multiobjective_algorithm_to_test: str,
      **kwargs):
    self._builtin_multiobjective_algorithm_to_test = (
        builtin_multiobjective_algorithm_to_test)
    super().__init__(backend_class, *args, **kwargs)

  def testSample(self):
    """Test `pg.sample` with Vizier backend."""
    self._backend_class.use_study_prefix('distributed_sampling')

    rewards_a = []
    algorithm = RandomAlgorithm()
    for a, fa in pg.sample(
        pg.Dict(x=pg.oneof([1, 2, 3])),  # Outer search space.
        algorithm=algorithm,
        num_examples=2,
        name='a',
    ):
      rs = []
      for b, fb in pg.sample(
          pg.Dict(y=pg.oneof([3, 4, 5, 6, 7])),  # Inner space.
          algorithm=backend.BuiltinAlgorithm('DEFAULT'),
          num_examples=3,
          name='a%d_b' % fa.id,
      ):
        logging.info('Sampling a=%s, fa=%s, b=%s, fb=%s', a, fa, b, fb)
        r = a.x + b.y
        fb(r)
        rs.append(r)
      ra = sum(rs) / len(rs)
      fa(ra)
      rewards_a.append(ra)

    # Check alogrithm state are updated.
    self.assertEqual(algorithm.num_proposals, 2)

    # Check studies are created.
    logging.info('Check studies.')
    self._backend_class._get_study_resource_name('distributed_sampling.a')
    self._backend_class._get_study_resource_name('distributed_sampling.a1_b')
    self._backend_class._get_study_resource_name('distributed_sampling.a2_b')

    # Check tuning results from each study.
    logging.info('Check tuning results from each study.')
    trials_a = pg.tuning.poll_result('a').trials
    self.assertLen(trials_a, 2)
    self.assertEqual([t.final_measurement.reward for t in trials_a], rewards_a)
    self.assertLen(pg.tuning.poll_result('a1_b').trials, 3)
    self.assertLen(pg.tuning.poll_result('a2_b').trials, 3)

    logging.info('Check that constant space raises an error.')
    # Non-deterministic value.
    with self.assertRaisesRegex(ValueError, "'space' is a constant value"):
      next(
          pg.sample(
              pg.Dict(x=1),  # a fixed value.
              algorithm=pg.geno.Random(seed=1),
              name='c',
          )
      )

  @absltest.skip('Pythia cannot accept empty trials.')
  def testSampleWithSearchAlgorithmStopIteration(self):
    """Test `pg.sample` with algorithm-side stop iteration."""
    self._backend_class.use_study_prefix('algorithm-stop-iteration')
    for sd, f in pg.sample(
        pg.Dict(x=pg.oneof(range(3)), y=pg.oneof(range(4))), pg.geno.Sweeping()
    ):
      f(sd.x + sd.y)
    self.assertLen(pg.tuning.poll_result('').trials, 3 * 4)

    # Continue from previous one.
    # We create a new Sweeping instance to test recovery of controller state.
    for _, _ in pg.sample(
        pg.Dict(x=pg.oneof(range(3)), y=pg.oneof(range(4))), pg.geno.Sweeping()
    ):
      assert False, 'should never happen'

    self.assertLen(pg.tuning.poll_result('').trials, 3 * 4)

  def testSampleWithRaceCondition(self):
    """Test `pg.sample` with race condition among co-workers."""
    self._backend_class.use_study_prefix('coworkers-with-race-conditions')
    _, f = next(pg.sample(pg.oneof([1, 2, 3]), pg.geno.Random(seed=1)))

    f(1)
    with self.assertRaisesRegex(
        pg.tuning.RaceConditionError,
        '.*Measurements can only be added to.*',
    ):
      f.add_measurement(0.1)

    with f.ignore_race_condition():
      f.add_measurement(0.1)

  def testSampleWithControllerSideUpdateOfDNAMetadata(self):
    """Test `pg.sample` with controller-side DNA metadata update."""
    self._backend_class.use_study_prefix('sampling_with_evolution')

    def write_metadata(dna_list):
      return [dna.set_metadata('foo', 1) for dna in dna_list]

    algorithm = pg.evolution.Evolution(
        (
            pg.evolution.Identity()[-1]
            >> pg.evolution.mutators.Uniform()
            >> write_metadata
        ),
        population_init=(pg.geno.Random(), 5),
    )

    for i, (x, fx) in enumerate(
        pg.sample(
            pg.oneof(range(100)),
            algorithm=algorithm,
            num_examples=10,
        )
    ):
      if i >= 5:
        self.assertEqual(fx.dna.metadata.foo, 1)
      fx(x)

    # Check alogrithm state are updated.
    self.assertEqual(algorithm.num_proposals, 10)

    # The feedback operation on the last trial is not yet called, since
    # it will be triggered by the next call to `GetNewSuggestions`.
    self.assertEqual(algorithm.num_feedbacks, 9)

    # Check studies are created.
    self._backend_class._get_study_resource_name('sampling_with_evolution')

    # Check the DNA metadata for each trial.
    trials = pg.tuning.poll_result('').trials
    self.assertLen(trials, 10)
    for i, t in enumerate(trials[:-1]):
      expected_metadata = {
          'proposal_id': i + 1,
          'generation_id': max(i - 5 + 1, 0) + 1,
          'feedback_sequence_number': i + 1,
          'initial_population': i < 5,
          'reward': float(t.dna.value),
      }
      if i >= 5:
        expected_metadata['foo'] = 1
      self.assertEqual(t.dna.metadata, expected_metadata)

    self.assertEqual(
        trials[-1].dna.metadata,
        {
            'proposal_id': 10,
            'generation_id': 6,
            'initial_population': False,
            'foo': 1,
        },
    )

  def testSamplingWithControllerSideEvaluation(self):
    """Test sampling with controller evaluated reward."""
    self._backend_class.use_study_prefix(
        'sampling_with_controller_side_evaluation'
    )

    def controller_side_evaluate(dna_list):
      return [pg.evolution.set_fitness(dna, 1.0) for dna in dna_list]

    algorithm = pg.evolution.Evolution(
        (
            pg.evolution.Identity()[-1]
            >> pg.evolution.mutators.Uniform()
            >> controller_side_evaluate
        ),
        population_init=(pg.geno.Sweeping(), 5),
    )

    client_side_evaluated = []
    for x, fx in pg.sample(
        pg.oneof(range(100)), algorithm=algorithm, num_examples=10
    ):
      client_side_evaluated.append(x)
      fx(x)

    # Check the DNA metadata for each trial.
    trials = pg.tuning.poll_result('').trials
    self.assertLen(trials, 10)
    for i, t in enumerate(trials[:-1]):
      self.assertEqual(
          t.dna.metadata,
          {
              'proposal_id': i + 1,
              'generation_id': max(i - 5 + 1, 0) + 1,
              'feedback_sequence_number': i + 1,
              'initial_population': i < 5,
              'reward': float(t.dna.value) if i < 5 else 1.0,
          },
          msg=f'for i={i}, t={t}',
      )

    # Check alogrithm state are updated.
    self.assertEqual(algorithm.num_proposals, 10)

    # The feedback operation on the last trial is not yet called, since
    # it will be triggered by the next call to `GetNewSuggestions`.
    self.assertEqual(algorithm.num_feedbacks, 9)

    # Check studies are created.
    self._backend_class._get_study_resource_name(
        'sampling_with_controller_side_evaluation'
    )

    # Only the initial population is evaluated at client side.
    self.assertEqual(client_side_evaluated, list(range(5)))

  def testSamplingWithMultiObjectiveAlgorithm(self):
    """Test sampling with multi-objective alogirthm."""
    # Test sampling with Vizier built-in algorithm.
    self._backend_class.use_study_prefix('multi-objective-sampling-vizier')
    sample1 = pg.sample(
        pg.oneof([1, 2]),
        backend.BuiltinAlgorithm(
            self._builtin_multiobjective_algorithm_to_test),
        metrics_to_optimize=['accuracy', 'latency'],
    )
    _, f = next(sample1)
    f(metrics={'accuracy': 0.9, 'latency': 0.5})

    class DummyMultiObjectiveAlgorithm(pg.DNAGenerator):
      """Test sampling with custom multi-objective algorithm."""

      @property
      def multi_objective(self):
        return True

      def setup(self, dna_spec):
        super().setup(dna_spec)
        self.rewards = []

      def _propose(self):
        return pg.DNA(1)

      def _feedback(self, dna, reward):
        self.rewards.append(reward)

    self._backend_class.use_study_prefix('multi-objective-sampling-custom')
    algo = DummyMultiObjectiveAlgorithm()
    sample2 = pg.sample(
        pg.oneof([1, 2]),
        algo,
        metrics_to_optimize=['reward', 'accuracy', 'latency'],
    )

    _, f = next(sample2)
    f(reward=0.0, metrics={'accuracy': 0.9, 'latency': 0.5})

    # In Vizier, feedback is done while the next example is requested.
    # Therefore, we call next sample to trigger the feedback call to the
    # first example.
    _, _ = next(sample2)
    self.assertEqual(algo.rewards, [(0.0, 0.9, 0.5)])

  def testSampleWithCustomTermination(self):
    """Test `pg.sample` with custom termination."""
    self._backend_class.use_study_prefix(None)
    hyper_value = pg.Dict(x=pg.oneof([1, 2, 3]))
    for x, f in pg.sample(
        hyper_value,
        algorithm=pg.geno.Random(seed=1),
        name='custom_termination',
    ):
      # Always invoke the feedback function in order to advance
      # to the next trial.
      if f.id == 1:
        f.skip()
      else:
        f.set_metadata('x', 'foo')
        f.set_metadata('y', True, per_trial=False)
        f.add_link('filepath', '/file/old_path')
        f(
            x.x,
            metrics={'accuracy': 0.9},
            metadata={'z': 'bar'},
            related_links={'filepath': '/file/path_%d' % f.id},
        )
      if f.id == 2:
        f.end_loop()
        logging.info('Ending loop')
        break

    result = pg.tuning.poll_result('custom_termination')
    self.assertFalse(result.is_active)
    self.assertEqual(result.metadata, {'y': True})
    self.assertLen(result.trials, 2)
    self.assertTrue(result.trials[0].infeasible)
    self.assertFalse(result.trials[1].infeasible)
    self.assertEqual(result.trials[1].measurements[0].reward, 3.0)
    self.assertEqual(
        result.trials[1].measurements[0].metrics,
        {
            # NOTE(daiyip): Vizier adds the reward to the metric with an empty
            # metric name.
            'reward': 3.0,
            'accuracy': 0.9,
        },
    )
    self.assertEqual(result.trials[1].metadata, {'x': 'foo', 'z': 'bar'})
    self.assertEqual(
        result.trials[1].related_links, {'filepath': '/file/path_2'}
    )
    self.assertEqual(
        str(result),
        inspect.cleandoc(
            """
        {
          'study': '%s',
          'status': {
            'COMPLETED': '2/2'
          },
          'best_trial': {
            'id': 2,
            'reward': 3.0,
            'step': 0,
            'dna': 'DNA(2)'
          }
        }"""
            % (
                self._backend_class._get_study_resource_name(
                    'custom_termination'
                )
            )
        ),
    )

    t = result.best_trial
    t.description = None
    self.assertEqual(
        str(t),
        inspect.cleandoc(
            """
        VizierTrial(
          id = 2,
          description = None,
          dna = DNA(
            value = 2,
            children = [],
            metadata = {}
          ),
          status = 'COMPLETED',
          final_measurement = Measurement(
            step = 0,
            elapse_secs = %s,
            reward = 3.0,
            metrics = {
              accuracy = 0.9,
              reward = 3.0
            },
            checkpoint_path = None
          ),
          infeasible = False,
          measurements = [
            0 : Measurement(
              step = 0,
              elapse_secs = %s,
              reward = 3.0,
              metrics = {
                accuracy = 0.9,
                reward = 3.0
              },
              checkpoint_path = None
            )
          ],
          metadata = {
            x = 'foo',
            z = 'bar'
          },
          related_links = {
            filepath = '/file/path_2'
          },
          created_time = %d,
          completed_time = %d
        )"""
            % (
                t.final_measurement.elapse_secs,
                t.measurements[0].elapse_secs,
                t.created_time,
                t.completed_time,
            )
        ),
    )

  def testSampleWithEarlyStopping(self):
    """Test sample with early stopping."""
    # Stop trial early if either the trial id or trial-step is greater than 1.
    early_stopping_policy = SizeLimitingStoppingPolicy(
        max_trial_id=1, max_step=1
    )
    actually_stopped = []
    for x, f in pg.sample(
        pg.oneof([1, 2, 3]),
        algorithm=RandomAlgorithm(),
        early_stopping_policy=early_stopping_policy,
        num_examples=2,
    ):
      skipped = False
      for i in range(4):
        f.add_measurement(x, step=i)
        # Add a minimal sleep to ensure that the measurement is added to DB.
        if f.should_stop_early():
          f.skip()
          skipped = True
          actually_stopped.append((f.id, i))
          break
      if not skipped:
        f.done()

    self.assertEqual(
        early_stopping_policy.requested_trial_steps,
        [(1, 0), (1, 1), (1, 2), (2, 0)],
    )
    self.assertEqual(
        early_stopping_policy.stopped_trial_steps, actually_stopped
    )

    self.assertEqual(
        early_stopping_policy.stopped_trial_steps, [(1, 2), (2, 0)]
    )

  def testSampleWithMultipleWorkers(self):
    """Test `pg.sample` with multiple workers."""
    self._backend_class.use_study_prefix(None)

    hyper_value = pg.Dict(x=pg.oneof([1, 2, 3]))

    # Create a new study for sample using Random.
    sample1 = pg.sample(
        hyper_value,
        algorithm=pg.geno.Random(seed=1),
        name='distributed_sampling2',
        group=0,
    )

    # `sample2` works on the same sampling queue with `sample1`
    # but with different trials by having the same `name` with
    # a different `group`.
    sample2 = pg.sample(
        hyper_value,
        algorithm=pg.geno.Random(seed=1),
        name='distributed_sampling2',
        group=1,
    )

    # `sample3` works on the same sampling queue and the same trials
    # with `sample1` by having the same `name` and `group`.
    sample3 = pg.sample(
        hyper_value,
        algorithm=pg.geno.Random(seed=1),
        name='distributed_sampling2',
        group=0,
    )

    # Make sure sampling with different worker IDs get different trial IDs.
    _, f1 = next(sample1)
    self.assertEqual(f1.id, 1)
    f1.set_metadata('x', 1)
    f1.set_metadata('y', RandomAlgorithm(), per_trial=False)

    # X is not serializable via `pg.to_json_str()`.
    class X:
      pass

    with self.assertRaisesRegex(
        ValueError, 'Cannot convert local class .* to JSON'
    ):
      f1.set_metadata('z', X)

    _, f2 = next(sample2)
    self.assertEqual(f2.id, 2)
    self.assertIsNone(f2.get_metadata('x'))
    self.assertEqual(f2.get_metadata('y', per_trial=False), RandomAlgorithm())

    _, f3 = next(sample3)
    self.assertEqual(f3.id, 1)
    self.assertEqual(f3.get_metadata('x'), 1)
    self.assertEqual(f3.get_metadata('y', per_trial=False), RandomAlgorithm())
    # Update the value of 'x'.
    f3.set_metadata('x', 2)
    f3.set_metadata('z', 'foo')
    f3.add_measurement(0.1, step=1)

    # Make sure sampling within the same worker get the same trial IDs before
    # feedback.
    _, f1b = next(sample1)
    self.assertEqual(f1b.id, 1)

    # Make sure `f1.trial` reflect the measurement added by f3.
    trial = f1.get_trial()
    self.assertLen(trial.measurements, 1)
    self.assertEqual(
        trial.measurements[0].reward,
        0.1,
        msg=f'Measurement was: {trial.measurements}',
    )

    logging.info('Trial metadata: %s', trial.metadata)
    self.assertEqual(
        trial.metadata.z, 'foo', msg=f'Metadata was: {trial.metadata}'
    )
    # Mark trial 1 as done.
    f1(0.5)
    # Check the value of metadata 'x' is updated (by f3).
    self.assertEqual(f1.get_metadata('x'), 2)
    self.assertEqual(f3.get_trial().status, 'COMPLETED')

    # NOTE(daiyip): trial 1 is now COMPLETED after `f1(0)`, therefore, calling
    # `f1b(reward)` will attempt to add a new measurement to the completed
    # trial, triggering a Vizier error.
    with self.assertRaisesRegex(
        pg.tuning.RaceConditionError,
        '.*Measurements can only be added to.*',
    ):
      f1b(1)

    # Make sure sampling within the same worker get different trial IDs after
    # previous trial is done.
    _, f1c = next(sample1)
    self.assertEqual(f1c.id, 3)
    _, f3b = next(sample3)
    self.assertEqual(f3b.id, 3)

    # After calling `end_loop`, all samplings (even with pending ones)
    # shall stop.
    f2.end_loop()
    with self.assertRaises(StopIteration):
      next(sample1)

    with self.assertRaises(StopIteration):
      next(sample2)

    with self.assertRaises(StopIteration):
      next(sample3)

  def testSampleWithDifferentSearchSpace(self):
    """Test client-side search space mismatch with server-side search space."""
    self._backend_class.use_study_prefix('different_search_space_vizier')

    ssd1 = pg.Dict(x=pg.oneof([1, 2, 3]))
    sample1 = pg.sample(ssd1, pg.geno.Random())
    x, f = next(sample1)
    f(x.x)

    ssd2 = pg.Dict(x=pg.oneof([1, 2]))
    sample2 = pg.sample(ssd2, pg.geno.Random())
    with self.assertRaisesRegex(ValueError, '.*different.*search space.*'):
      next(sample2)

  def testSampleWithCustomHyper(self):
    """Test sample with custom hyper."""
    self._backend_class.use_study_prefix('custom_hyper_vizier')

    class VariableString(pg.hyper.CustomHyper):
      """A custom decision type that represents a variable-length string."""

      def custom_decode(self, dna):
        return dna.value

    @pg.geno.dna_generator
    def init_population(dna_spec):
      yield pg.DNA('abc', spec=dna_spec)

    class StringRepeater(pg.evolution.Mutator):

      def mutate(self, dna):  # type: ignore
        return pg.DNA(dna.value + dna.value, spec=dna.spec)

    algo = pg.evolution.Evolution(
        StringRepeater(),
        population_init=init_population.partial(),
        population_update=pg.evolution.selectors.Last(1),
    )

    sample = pg.sample(VariableString(), algo)
    x, f = next(sample)
    self.assertEqual(x, 'abc')
    f(len(x))

    x, f = next(sample)
    self.assertEqual(x, 'abcabc')
    f(len(x))

    x, f = next(sample)
    self.assertEqual(x, 'abcabcabcabc')
    f(len(x))

  def testAllProposedTrialAreDeliveredToWorkers(self):
    """Test that all proposed trials are delivered to the workers."""
    self._backend_class.use_study_prefix('proposal_delivery')
    searchable_list = pg.List(
        [
            pg.one_of([1, 2, 3]),
            pg.one_of([-1, 0, 1]),
            pg.one_of([1, 2, 0]),
        ]
        * 5
    )
    algorithm = pg.evolution.regularized_evolution()

    def worker_fun():
      for l, f in pg.sample(searchable_list, algorithm, num_examples=100):
        reward = float(sum(l))
        f(reward)

    workers = [threading.Thread(target=worker_fun) for _ in range(10)]
    for w in workers:
      w.start()
    for w in workers:
      w.join()

    result = pg.tuning.poll_result('')
    for t in result.trials:
      self.assertEqual(t.id, t.dna.metadata.proposal_id)



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""List of Designers and related algorithms."""

from vizier._src.algorithms.designers.bocs import BOCSDesigner
from vizier._src.algorithms.designers.cmaes import CMAESDesigner
from vizier._src.algorithms.designers.eagle_strategy.eagle_strategy import EagleStrategyDesigner
from vizier._src.algorithms.designers.gp_bandit import VizierGPBandit
from vizier._src.algorithms.designers.gp_ucb_pe import VizierGPUCBPEBandit
from vizier._src.algorithms.designers.grid import GridSearchDesigner
from vizier._src.algorithms.designers.harmonica import HarmonicaDesigner
from vizier._src.algorithms.designers.quasi_random import QuasiRandomDesigner
from vizier._src.algorithms.designers.random import RandomDesigner
from vizier._src.algorithms.designers.unsafe_as_infeasible_designer import UnsafeAsInfeasibleDesigner
from vizier._src.algorithms.ensemble.ensemble_design import AdaptiveEnsembleDesign
from vizier._src.algorithms.ensemble.ensemble_design import EnsembleDesign
from vizier._src.algorithms.ensemble.ensemble_design import EXP3IXEnsembleDesign
from vizier._src.algorithms.ensemble.ensemble_design import EXP3UniformEnsembleDesign
from vizier._src.algorithms.ensemble.ensemble_design import RandomEnsembleDesign
from vizier._src.algorithms.evolution.nsga2 import NSGA2Designer



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Experimenter that normalizes the range of each metric."""

import collections
import copy
from typing import Dict, Sequence

import numpy as np
from vizier import pyvizier as vz
from vizier._src.benchmarks.experimenters import experimenter
from vizier.pyvizier import converters


# TODO: Allow more sampling methods.
class NormalizingExperimenter(experimenter.Experimenter):
  """Normalizes an Experimenter output via (y - empirical_mean) / empirical_std."""

  def __init__(
      self,
      exptr: experimenter.Experimenter,
      num_normalization_samples: int = 100,
      noise_seed: int = 42,
  ):
    """Normalizing experimenter uses a grid to estimate a normalization constant.

    Args:
      exptr: Experimenter to be normalized.
      num_normalization_samples: Number of samples to determine normalization.
      noise_seed: Random seed for tiebreaking categorical samples.
    """
    self._exptr = exptr
    self._problem_statement = exptr.problem_statement()

    # Convert search space into hypercube, then sample.
    converter = converters.TrialToArrayConverter.from_study_config(
        study_config=self._problem_statement, scale=True, pad_oovs=False
    )
    feature_dim = sum([o.num_dimensions for o in converter.output_specs])

    normalized_samples = np.linspace(
        np.zeros(feature_dim), np.ones(feature_dim), num_normalization_samples
    )

    # Categoricals use argmax and break ties by always choosing the first
    # feasible choice. Let's randomize the tiebreaking.
    noise = np.random.RandomState(noise_seed).normal(
        scale=1e-6, size=(num_normalization_samples, feature_dim)
    )
    normalized_samples += noise

    sampled_params = converter.to_parameters(normalized_samples)
    metrics = collections.defaultdict(list)
    for parameters in sampled_params:
      trial = vz.Trial(parameters=parameters)
      exptr.evaluate([trial])
      measurement = trial.final_measurement
      for name, metric in (measurement.metrics if measurement else {}).items():
        metrics[name].append(metric.value)

    self._norm_means: Dict[str, float] = {}
    self._norm_stds: Dict[str, float] = {}
    for name, grid_values in metrics.items():
      self._norm_means[name] = np.mean(grid_values)
      self._norm_stds[name] = np.std(grid_values) + 1e-7

  def problem_statement(self) -> vz.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def evaluate(self, suggestions: Sequence[vz.Trial]):
    self._exptr.evaluate(suggestions)
    for suggestion in suggestions:
      if suggestion.final_measurement is None:
        continue
      normalized_metrics: Dict[str, vz.Metric] = {}
      for name, metric in suggestion.final_measurement.metrics.items():
        norm_val = metric.value - self._norm_means[name]
        norm_val /= self._norm_stds[name]
        normalized_metrics[name] = vz.Metric(norm_val)
      suggestion.final_measurement.metrics = normalized_metrics

  def __repr__(self):
    return (
        'NormalizingExperimenter with normalization means'
        f' {self._norm_means} and stds {self._norm_stds} on {self._exptr}'
    )


class HyperCubeExperimenter(experimenter.Experimenter):
  """Normalizes the search space into unit hypercube using converter."""

  def __init__(self, exptr: experimenter.Experimenter):
    self._exptr = exptr
    original_problem = exptr.problem_statement()

    converter = converters.TrialToArrayConverter.from_study_config(
        study_config=original_problem, scale=True, pad_oovs=False
    )
    self._converter = converter
    feature_dim = sum([o.num_dimensions for o in converter.output_specs])

    new_space = vz.SearchSpace()  # Setup hypercube search space.
    for i in range(feature_dim):
      new_space.add(vz.ParameterConfig.factory(f'h{i}', bounds=(0.0, 1.0)))
    self._problem_statement = copy.deepcopy(original_problem)
    self._problem_statement.search_space = new_space

    # For simply converting new hypercube parameters to numpy arrays.
    self._eval_converter = converters.TrialToArrayConverter.from_study_config(
        study_config=self._problem_statement, scale=False, pad_oovs=False
    )

  def problem_statement(self) -> vz.ProblemStatement:
    return copy.deepcopy(self._problem_statement)

  def evaluate(self, suggestions: Sequence[vz.Trial]):
    hypercube_features = self._eval_converter.to_features(suggestions)
    original_params = self._converter.to_parameters(hypercube_features)

    orig_suggestions = copy.deepcopy(suggestions)
    for param_dict, orig_suggestion in zip(original_params, orig_suggestions):
      orig_suggestion.parameters = param_dict

    self._exptr.evaluate(orig_suggestions)

    for suggestion, orig_suggestion in zip(suggestions, orig_suggestions):
      suggestion.final_measurement = orig_suggestion.final_measurement



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

from vizier import pyvizier
from vizier._src.benchmarks.experimenters import combo_experimenter
from vizier._src.benchmarks.experimenters import normalizing_experimenter
from vizier._src.benchmarks.experimenters import numpy_experimenter
from vizier._src.benchmarks.experimenters.synthetic import bbob
from vizier._src.benchmarks.experimenters.synthetic import multiarm

from absl.testing import absltest
from absl.testing import parameterized


class NormalizingExperimenterTest(parameterized.TestCase):

  @parameterized.named_parameters(
      ('Sphere', bbob.Sphere),
      ('Rastrigin', bbob.Rastrigin),
      ('BuecheRastrigin', bbob.BuecheRastrigin),
      ('LinearSlope', bbob.LinearSlope),
      ('AttractiveSector', bbob.AttractiveSector),
      ('StepEllipsoidal', bbob.StepEllipsoidal),
      ('RosenbrockRotated', bbob.RosenbrockRotated),
      ('Discus', bbob.Discus),
      ('BentCigar', bbob.BentCigar),
      ('SharpRidge', bbob.SharpRidge),
      ('DifferentPowers', bbob.DifferentPowers),
      ('Weierstrass', bbob.Weierstrass),
      ('SchaffersF7', bbob.SchaffersF7),
      ('SchaffersF7IllConditioned', bbob.SchaffersF7IllConditioned),
      ('GriewankRosenbrock', bbob.GriewankRosenbrock),
      ('Schwefel', bbob.Schwefel),
      ('Katsuura', bbob.Katsuura),
      ('Lunacek', bbob.Lunacek),
      ('Gallagher101Me', bbob.Gallagher101Me),
  )
  def testNormalizationApply(self, func):
    dim = 5
    exptr = numpy_experimenter.NumpyExperimenter(
        func, bbob.DefaultBBOBProblemStatement(dim)
    )
    normalizing_exptr = normalizing_experimenter.NormalizingExperimenter(
        exptr=exptr
    )

    parameters = exptr.problem_statement().search_space.parameters
    self.assertLen(parameters, dim)

    t = pyvizier.Trial(
        parameters={
            param.name: float(index) for index, param in enumerate(parameters)
        }
    )

    exptr.evaluate([t])
    metric_name = exptr.problem_statement().metric_information.item().name

    normalizing_exptr.evaluate([t])
    normalized_value = t.final_measurement_or_die.metrics[metric_name].value
    self.assertBetween(normalized_value, -10, 10)

  def test_NormalizingCategoricals(self):
    mab_exptr = multiarm.FixedMultiArmExperimenter(
        arms_to_rewards={'0': -1e6, '1': 0.0, '2': 1e6}
    )
    norm_exptr = normalizing_experimenter.NormalizingExperimenter(mab_exptr)
    metric_name = norm_exptr.problem_statement().metric_information.item().name

    for arm in range(3):
      t = pyvizier.Trial(parameters={'arm': str(arm)})
      norm_exptr.evaluate([t])
      normalized_value = t.final_measurement_or_die.metrics[metric_name].value
      self.assertBetween(normalized_value, -10, 10)


class HyperCubeExperimenterTest(parameterized.TestCase):

  def testE2E(self):
    num_boolean_params = 3

    original_exptr = combo_experimenter.ContaminationExperimenter(
        contamination_n_stages=num_boolean_params
    )
    hypercube_exptr = normalizing_experimenter.HyperCubeExperimenter(
        original_exptr
    )

    problem = hypercube_exptr.problem_statement()
    metric_name = problem.metric_information.item().name

    # 3 Booleans -> 6 total coordinates after one-hots.
    self.assertLen(problem.search_space.parameters, 2 * num_boolean_params)

    t = pyvizier.Trial(
        parameters={
            'h0': 0.0,
            'h1': 1.0,
            'h2': 1.0,
            'h3': 0.0,
            'h4': 0.3,
            'h5': 0.6,
        }
    )
    hypercube_exptr.evaluate([t])
    hypercube_value = t.final_measurement_or_die.metrics[metric_name].value

    # Argmax over hypercube one-hots.
    t = pyvizier.Trial(
        parameters={'x_0': 'True', 'x_1': 'False', 'x_2': 'True'}
    )
    original_exptr.evaluate([t])
    original_value = t.final_measurement_or_die.metrics[metric_name].value

    self.assertEqual(hypercube_value, original_value)


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""A helper functionality to handle singleton parameters."""

import copy
import logging
from typing import Sequence

import attrs
from vizier import pyvizier as vz


@attrs.define
class SingletonParameterHandler:
  """A helper class to handle singleton parameters.

  The class allows to remove singleton parameters (i.e. having a single value)
  from the problem statement's search space and from trial, so that designers
  won't have to handle them. In addition, it allows to re-add the singleton
  values back to the trial suggestions, so that the users can access them if
  needed.
  """

  problem: vz.ProblemStatement
  # ----------------------------------------------------------------------------
  # Internal attributes
  # ----------------------------------------------------------------------------
  _singletons: dict[str, vz.ParameterValueTypes] = attrs.field(init=False)
  _stripped_problem: vz.ProblemStatement = attrs.field(init=False)

  @property
  def stripped_problem(self) -> vz.ProblemStatement:
    """Returns the stripped problem."""
    return self._stripped_problem

  def __attrs_post_init__(self):
    logging.info("problem: %s", self.problem)
    self._singletons = self._find_singletons()
    self._stripped_problem = self._strip_problem()

  def _find_singletons(self) -> dict[str, vz.ParameterValueTypes]:
    """Finds the singleton parameters in the problem."""
    singletons = {}
    for param in self.problem.search_space.parameters:
      if param.type == vz.ParameterType.DOUBLE:
        if param.bounds[0] == param.bounds[1]:
          singletons[param.name] = param.bounds[0]
      elif param.type == vz.ParameterType.INTEGER:
        if param.bounds[0] == param.bounds[1]:
          singletons[param.name] = param.bounds[0]
      elif param.type in (
          vz.ParameterType.CATEGORICAL,
          vz.ParameterType.DISCRETE,
      ):
        if len(param.feasible_values) == 1:
          singletons[param.name] = param.feasible_values[0]
      elif param.type == vz.ParameterType.CUSTOM:
        pass
      else:
        raise ValueError("Unknown parameter type: %s" % param.type)
    return singletons

  def _strip_problem(self) -> vz.ProblemStatement:
    """Strips the problem of the singleton parameters."""
    stripped_problem = copy.deepcopy(self.problem)
    for param in self.problem.search_space.parameters:
      if param.name in self._singletons:
        stripped_problem.search_space.pop(param.name)
    return stripped_problem

  def strip_trials(
      self, trials: Sequence[vz.TrialSuggestion]
  ) -> Sequence[vz.TrialSuggestion]:
    """Strips the trials of the singleton parameters."""
    if not self._singletons:
      return trials
    new_trials = []
    for trial in trials:
      new_trial = copy.deepcopy(trial)
      for param_name in trial.parameters:
        if param_name in self._singletons:
          del new_trial.parameters[param_name]
      new_trials.append(new_trial)
    return new_trials

  def augment_trials(
      self, trials: Sequence[vz.TrialSuggestion]
  ) -> Sequence[vz.TrialSuggestion]:
    """Augments the trials with the singleton parameters."""
    if not self._singletons:
      return trials
    new_trials = []
    for trial in trials:
      new_trial = copy.deepcopy(trial)
      for singleton_name, singleton_value in self._singletons.items():
        new_trial.parameters[singleton_name] = singleton_value
      new_trials.append(new_trial)
    return new_trials



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for problem statement converter."""

import copy

import numpy as np
from vizier import pyvizier as vz
from vizier.pyvizier.converters import embedder

from absl.testing import absltest


class ProblemStatementTest(absltest.TestCase):

  def test_convert_search_space(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_int_param('i1', 0, 10)
    problem.search_space.root.add_discrete_param('d1', [0, 10, 20])
    problem.search_space.root.add_categorical_param('c1', ['a', 'b', 'c'])
    converter = embedder.ProblemAndTrialsScaler(problem)

    new_space = converter.problem_statement.search_space
    self.assertEqual(new_space.get('f1').type, vz.ParameterType.DOUBLE)
    self.assertEqual(new_space.get('f1').bounds, (0.0, 1.0))

    self.assertEqual(new_space.get('i1').type, vz.ParameterType.DOUBLE)
    self.assertEqual(new_space.get('i1').bounds, (0.0, 1.0))

    self.assertEqual(new_space.get('d1').type, vz.ParameterType.DISCRETE)
    self.assertEqual(new_space.get('d1').feasible_values, [0.0, 0.5, 1.0])

    self.assertEqual(new_space.get('c1').type, vz.ParameterType.CATEGORICAL)
    self.assertEqual(new_space.get('c1').feasible_values, ['a', 'b', 'c'])

  def test_map_trials(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_int_param('i1', 0, 10)
    problem.search_space.root.add_discrete_param('d1', [0, 10, 20])
    problem.search_space.root.add_categorical_param('c1', ['a', 'b', 'c'])
    converter = embedder.ProblemAndTrialsScaler(problem)

    trial1 = vz.Trial(parameters={'f1': 1.0, 'i1': 0, 'd1': 10, 'c1': 'b'})
    trial2 = vz.Trial(parameters={'f1': 4.0, 'i1': 8, 'd1': 0, 'c1': 'a'})
    trial1.final_measurement = vz.Measurement(metrics={'o': 1.0})
    trial2.final_measurement = vz.Measurement(metrics={'o': 2.0})
    suggestion_trials = [trial1, trial2]
    trial1_copy = copy.deepcopy(trial1)
    trial2_copy = copy.deepcopy(trial2)

    mapped_trials = converter.map(suggestion_trials)
    self.assertAlmostEqual(mapped_trials[0].parameters['f1'].value, 0.1)
    self.assertAlmostEqual(mapped_trials[0].parameters['i1'].value, 0.0)
    self.assertAlmostEqual(mapped_trials[0].parameters['d1'].value, 0.5)
    self.assertEqual(mapped_trials[0].parameters['c1'].value, 'b')
    # Check that the trial measurement remains the same.
    self.assertEqual(mapped_trials[0].final_measurement.metrics['o'].value, 1.0)
    self.assertAlmostEqual(mapped_trials[1].parameters['f1'].value, 0.4)
    self.assertAlmostEqual(mapped_trials[1].parameters['i1'].value, 0.8)
    self.assertAlmostEqual(mapped_trials[1].parameters['d1'].value, 0.0)
    self.assertEqual(mapped_trials[1].parameters['c1'].value, 'a')
    # Check that the trial measurement remains the same.
    self.assertEqual(mapped_trials[1].final_measurement.metrics['o'].value, 2.0)
    # Check that the original trials weren't change during map.
    self.assertEqual(trial1, trial1_copy)
    self.assertEqual(trial2, trial2_copy)

  def test_unmap_trials(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param('f1', 0.0, 10.0)
    problem.search_space.root.add_int_param('i1', 0, 10)
    problem.search_space.root.add_discrete_param('d1', [0, 10, 20])
    problem.search_space.root.add_categorical_param('c1', ['a', 'b', 'c'])
    converter = embedder.ProblemAndTrialsScaler(problem)

    trial1 = vz.Trial(parameters={'f1': 1.0, 'i1': 0, 'd1': 10, 'c1': 'b'})
    trial2 = vz.Trial(parameters={'f1': 4.0, 'i1': 8, 'd1': 0, 'c1': 'c'})
    trial3 = vz.Trial(parameters={'f1': 0.0, 'i1': 8, 'd1': 20, 'c1': 'b'})
    trial4 = vz.Trial(parameters={'f1': 10.0, 'i1': 2, 'd1': 10, 'c1': 'a'})
    suggestion_trials = [trial1, trial2, trial3, trial4]
    suggestion_trials_copy = copy.deepcopy(suggestion_trials)
    recovered_trials = converter.unmap(converter.map(suggestion_trials))

    for i, trial in enumerate(suggestion_trials):
      for param_config in problem.search_space.parameters:
        name = param_config.name
        if param_config.type == vz.ParameterType.CATEGORICAL:
          self.assertEqual(
              trial.parameters[name].value,
              recovered_trials[i].parameters[name].value,
          )
        else:
          self.assertAlmostEqual(
              trial.parameters[name].value,
              recovered_trials[i].parameters[name].value,
              places=5,
          )
    # Check that the original trials weren't change during unmap.
    for i in range(len(suggestion_trials)):
      self.assertEqual(suggestion_trials_copy[i], suggestion_trials[i])

  def test_non_linear_scale(self):
    problem = vz.ProblemStatement()
    problem.search_space.root.add_float_param(
        'f1', 1.0, 100.0, scale_type=vz.ScaleType.LOG
    )
    trial = vz.Trial(parameters={'f1': 10.0})
    converter = embedder.ProblemAndTrialsScaler(problem)
    new_space = converter.problem_statement.search_space
    self.assertEqual(new_space.get('f1').type, vz.ParameterType.DOUBLE)
    self.assertEqual(new_space.get('f1').bounds, (0.0, 1.0))
    mapped_trial = converter.map([trial])[0]
    self.assertAlmostEqual(
        mapped_trial.parameters['f1'].value, np.log(10.0) / np.log(100.0)
    )
    unmapped_trial = converter.unmap([mapped_trial])[0]
    self.assertAlmostEqual(
        unmapped_trial.parameters['f1'].value, 10.0, places=5
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Large-scale stress tests (multiple clients, mulithreading, etc.) for Vizier Service."""

import multiprocessing.pool
import time

from absl import logging
from vizier._src.service import constants
from vizier._src.service import vizier_client
from vizier._src.service import vizier_server
from vizier.benchmarks import experimenters
from vizier.service import pyvizier

from absl.testing import absltest
from absl.testing import parameterized


class PerformanceTest(parameterized.TestCase):
  server: vizier_server.DefaultVizierServer

  @classmethod
  def setUpClass(cls):
    super().setUpClass()
    cls.server = vizier_server.DefaultVizierServer(
        database_url=constants.SQL_MEMORY_URL
    )
    vizier_client.environment_variables.server_endpoint = cls.server.endpoint

  @parameterized.parameters((1, 10), (2, 10), (10, 10), (50, 5), (100, 5))
  def test_multiple_clients_basic(
      self, num_simultaneous_clients, num_trials_per_client
  ):
    def fn(client_id: int):
      experimenter = experimenters.BBOBExperimenterFactory('Sphere', 2)()
      problem_statement = experimenter.problem_statement()
      study_config = pyvizier.StudyConfig.from_problem(problem_statement)
      study_config.algorithm = pyvizier.Algorithm.RANDOM_SEARCH

      client = vizier_client.create_or_load_study(
          owner_id='my_username',
          study_id=self.id(),  # Use the testcase name.
          study_config=study_config,
          client_id=str(client_id),
      )

      for _ in range(num_trials_per_client):
        suggestions = client.get_suggestions(suggestion_count=1)
        experimenter.evaluate(suggestions)
        for completed_trial in suggestions:
          client.complete_trial(
              completed_trial.id, completed_trial.final_measurement
          )

      return client

    client_ids = range(num_simultaneous_clients)
    pool = multiprocessing.pool.ThreadPool(num_simultaneous_clients)

    start = time.time()
    clients = pool.map(fn, client_ids)
    end = time.time()
    pool.close()

    self.assertEqual(
        max({t.id for t in clients[0].list_trials()}),
        num_simultaneous_clients * num_trials_per_client,
    )

    logging.info(
        'For %d clients to evaluate %d trials each, it took %f seconds total.',
        num_simultaneous_clients,
        num_trials_per_client,
        end - start,
    )


if __name__ == '__main__':
  absltest.main()



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Shared classes for representing studies."""

from typing import List
import attr
from vizier._src.pyvizier.shared import base_study_config
from vizier._src.pyvizier.shared import trial


@attr.define(frozen=True, init=True, slots=True, kw_only=False)
class ProblemAndTrials:
  """Container for problem statement and trials."""
  problem: base_study_config.ProblemStatement = attr.ib(init=True)
  trials: List[trial.Trial] = attr.ib(
      init=True,
      # TODO: Remove the pylint.
      converter=lambda x: list(x),  # pylint: disable=unnecessary-lambda
      default=attr.Factory(list))



# Copyright 2024 Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations

"""Tests for bocs."""
from vizier._src.algorithms.designers import bocs
from vizier._src.algorithms.testing import test_runners
from vizier._src.benchmarks.experimenters import combo_experimenter

from absl.testing import absltest
from absl.testing import parameterized


class BocsTest(parameterized.TestCase):

  @parameterized.parameters((bocs.SemiDefiniteProgramming,),
                            (bocs.SimulatedAnnealing,))
  def test_make_suggestions(self, acquisition_optimizer_factory):
    experimenter = combo_experimenter.IsingExperimenter(lamda=0.01)
    designer = bocs.BOCSDesigner(
        experimenter.problem_statement(),
        acquisition_optimizer_factory=acquisition_optimizer_factory,
        num_initial_randoms=1)

    trials = test_runners.run_with_random_metrics(
        designer,
        experimenter.problem_statement(),
        iters=5,
        batch_size=1,
        verbose=1,
        validate_parameters=True)
    self.assertLen(trials, 5)


if __name__ == '__main__':
  absltest.main()
